/*
** This program is part of the FLoP project
**
** Copyright (C) 2002-2004 Dirk Geschke <Dirk@geschke-online.de>
**
** This program is free software; you can redistribute it and/or modify
** it under the terms of the GNU General Public License as published by
** the Free Software Foundation; either version 2 of the License, or
** (at your option) any later version.
**
** This program is distributed in the hope that it will be useful,
** but WITHOUT ANY WARRANTY; without even the implied warranty of
** MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
** GNU General Public License for more details.
**
** You should have received a copy of the GNU General Public License
** along with this program; if not, write to the Free Software
** Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

/*************************************************************************/
/*                                                                       */
/* Program to receive network data from a snort database based on the    */
/* sensor id (sid) and counter id (cid). From this data a network packet */
/* in pcap format is written to a file or stdout (filename = '-'). This  */
/* file can be used to analyze it via tcpdump or ethereal to use their   */
/* protocol analyzing features. Simply to view the ascii payload may not */
/* be enough.                                                            */
/*                                                                       */
/* Note: You need an extended database scheme in contrast to the normal  */
/*       snort database used in conjunction with acid                    */
/*                                                                       */
/*************************************************************************/

static char *rcsid="$Id: getpacket.c,v 1.2 2004/03/30 21:38:57 geschke Exp $";

#include "FLoPconfig.h"
#undef PACKAGE
#undef VERSION
#undef PACKAGE
#undef VERSION
#undef PACKAGE_BUGREPORT
#undef PACKAGE_NAME
#undef PACKAGE_STRING
#undef PACKAGE_TARNAME
#undef PACKAGE_VERSION

#ifndef HAVE_U_INT8_T
#define u_int8_t  uint8_t
#endif
#ifndef HAVE_U_INT16_T
#define u_int16_t uint16_t
#endif
#ifndef HAVE_U_INT32_T
#define u_int32_t uint32_t
#endif

#include <sys/types.h>
#include <stdio.h>
#include <sys/un.h>
#include <errno.h>
#include <unistd.h>

#include "snort.h"
#include "plugbase.h"

#define STRLEN 256
#define MAX_QUERY_LENGTH 8192
#define MAXBUFFSIZE 65536 /* at least one complete TCP session */

#ifndef UNIX_PATH_MAX
#define UNIX_PATH_MAX   108
#endif

#ifdef MYSQL_DB
#include <mysql/mysql.h>
#endif

#ifdef POSTGRES
#include <libpq-fe.h>
#endif

#define DB_MYSQL 1
#define DB_POSTGRES 2


typedef struct _GPoptions
{
  int32_t sid;
  int32_t cid;
  char ConfigFile[STRLEN];
  char DumpFile[STRLEN];
  char DBuser[STRLEN];
  char DBname[STRLEN];
  int DBtype;
  int DBencoding;
  char DBpassword[STRLEN];
  char SocketName[STRLEN];
#ifdef MYSQL_DB
    MYSQL * m_sock;
    MYSQL_RES * m_result;
    MYSQL_ROW m_row;
#endif
#ifdef POSTGRES
   PGconn     *p_sock;
   PGresult   *p_result;
#endif
} GPoptions;


/**** GLOBAL VALUES ***/

static char *select_ptr=NULL; /* Memory is allocted in DatabaseInit! */
static char *insert_ptr=NULL; /* Memory is allocted in DatabaseInit! */
static char *dbuffer=NULL;    /* Needed in Database() */
static char *tbuffer=NULL;  /* Needed in Database() to store palyoad */
static char *pbuffer=NULL;  /* Needed in Database() to store palyoad */


/***** PROTOTYPES *****/

void CmdLine(int argc, char* argv[], GPoptions *opt);
void Usage(char *argv);
int Select(char * query, GPoptions * data);
char * Select1(char * query, GPoptions * data);
int Insert(char * query, GPoptions * data);
int Connect(GPoptions * data);
void Disconnect(GPoptions * data);
int DatabaseInit(GPoptions * data);
void ReadConfig(GPoptions* Options);
void ArgCheck(char *arg);
int GetPacket(GPoptions *Options);
int BuildPacket(GPoptions* Options);


/****** MAIN *****/

int 
main(int argc, char* argv[])
{
  GPoptions Options;
  int res;

  bzero(&Options,sizeof(Options));

  /* preset some values */
  snprintf(Options.DumpFile,STRLEN,"/var/tmp/dump");
  snprintf(Options.ConfigFile,STRLEN,"%s/getpacket.conf",PATH_CONFIG);

  /* get the command line arguments */
  CmdLine(argc,argv,&Options);

#ifdef DEBUG
  printf("sid: %i, cid: %i\n",Options.sid,Options.cid);
#endif

  /* Check minimal necessary values */
  if (Options.sid < 1)
    {
      fprintf(stderr,"You have to set SID to a positive value: %i\n",
	      Options.sid);
      fprintf(stderr,"Use option -S to set the SID.\n\n");
      exit(-1);
    }

  if (Options.cid < 1)
    {
      fprintf(stderr, "You have to set CID to a positive value: %i\n",
	      Options.cid);
      fprintf(stderr, "Use option -c to set the CID.\n\n");
      exit(-1);
    }

  /* read in the database options from a file, the servsock.conf file
   * will work here also, additional options will be ignored.
   */
  ReadConfig(&Options);

  /* Special case for PostgreSQL, if socket is "NULL" set it to real
   * zero value
   */
  if (!strncmp(Options.SocketName,"NULL",4)) Options.SocketName[0]='\0';

  /* Open database connection and do some checks */
  if(res=DatabaseInit(&Options)) 
  {
    if (res!=2) Disconnect(&Options);
    exit(1);
  };

  /* fetch the payload information based on sid and cid from database */
  if(GetPacket(&Options))
    {
      fprintf(stderr,"An error occured in GetPacket()-> Exiting.\n");
      Disconnect(&Options);
      exit(1);
    }

  /* At this point we have all data from the database */
  Disconnect(&Options);

#ifdef DEBUG
  printf("pcap_header : <%s>\n",pbuffer);
  printf("data_header : <%s>\n",tbuffer);
  printf("data_payload: <%s>\n",dbuffer);
#endif

  /* finally build the pcap packet and write it to a file or stdout */
  if( BuildPacket(&Options))
    {
      fprintf(stderr, "An error occured in BuildPacket-> Exiting.\n");
      exit(1);
    }

  exit(0);
}


/* 
 * Function: CmdLine
 *
 * Purpose: Set or modify some values needed for startup
 *
 * Arguments: Commandline argv and number of elements argc
 *
 * Returns: Nothing, we set only global values
 *
 * Author: DG
 *
 */

void 
CmdLine(int argc, char* argv[], GPoptions *opt) 
{
  int o, size, i;
  char options[]="hvw:c:C:S:";

  while((o = getopt(argc, argv, options)) != -1) 
    {
      switch(o)
	{
	case 'v':
	  printf("\nProgram to extract full packet data from the database\n");
	  printf(" and create a pcap file to analyse it with programs like\n");
          printf(" tcpdump or ethereal.\n");
	  printf("\nWritten by Dirk Geschke\n\n");
	  printf("%s\n\n",rcsid);
	  printf("Version %s (snort-%s)\n\n",FLoP_VERSION,VERSION);
	  exit(0);
	  break;
      
	case 'h':
	  Usage(argv[0]);
	  break;
	  
	case 'c':
	 size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	 strncpy(opt->ConfigFile,optarg, size);
	 opt->ConfigFile[size]='\0';
#ifdef DEBUG_CMDLINE
	 printf("ConfigFile: %s\n",opt->ConfigFile); 
#endif
	  break;
	  
	case 'C':
	  opt->cid=atoi(optarg);
#ifdef DEBUG_CMDLINE
	  printf("Packet-ID: %s\n",opt->cid);
#endif
	  break;  

	case 'S':
	  opt->sid=atoi(optarg);
#ifdef DEBUG_CMDLINE
	  printf("Sensor-ID: %s\n",opt->sid);
#endif
	  break;  

        case 'd':
          //DumpConf=1;
#ifdef DEBUG_CMDLINE
          printf("Dump configuration.\n");
#endif
          break;

	case 'w':
	  size=strlen(optarg)< STRLEN ? strlen(optarg) : STRLEN-1;
	  strncpy(opt->DumpFile,optarg, size);
	  opt->DumpFile[size]='\0';
#ifdef DEBUG_CMDLINE
	  printf("Working directory for daemon mode: %s\n",opt->DumpFile);
#endif
	  break;  

	default:
	  fprintf(stderr,"Unknown flag: %c\n",opt);
	  Usage(argv[0]);
	}
    }
}

/* 
 * Function: Usage
 *
 * Purpose: Display the possible command line arguments
 *
 * Arguments: argv, only needed to know the program name
 *
 * Returns: No return, simply exit...
 *
 * Author: DG
 *
 */

void 
Usage(char *argv)
{
  printf("\nUsage: %s [-hv] [-c ConfigFile] -C PacketCount"
	 " -S Sensor-ID [-w DumpFile] \n",argv);
  printf("\n -c ConfigFile : Configuration file for database access");
  printf("\n -C PacketCount: Counter of packet related to sensor (cid)");
  printf("\n -S Sensor-ID  : To which sensor belongs the PacketCount");
  printf("\n -w DumpFile   : Where should the data be written to");
  printf("\n\n");
  exit(0);
}

/*
 * Function: DatabaseInit(GPoptions *)
 *
 * Purpose: Connects to database, sets/gets the sensor id/name/encoding
 *          and selects the starting cid from database. Should we change
 *          this to scheme v106?
 *
 * Arguments: args => ptr to database parameters
 *
 * Returns: 0 on success
 *
 */

int 
DatabaseInit(GPoptions *data)
{

  /* allocate memory for configuration queries */
  if (select_ptr==NULL)
    {
      select_ptr = (char *) malloc(MAX_QUERY_LENGTH+1);
      insert_ptr = (char *) malloc(MAX_QUERY_LENGTH+1);
    }
  if ( (select_ptr==NULL) || (insert_ptr==NULL) )
	{
	  fprintf(stderr,"DatabaseInit: problem to allocate memory -> Exiting.\n");
	  return(1);
	}

  /* These values are needed for the Database() function */
  if (pbuffer==NULL)
    {
      dbuffer = (char *) malloc (MAX_QUERY_LENGTH+1);
      tbuffer = (char *) malloc ((MAXBUFFSIZE*2)+1);
      pbuffer = (char *) malloc ((sizeof(struct pcap_pkthdr)*2)+1);
    }
  if ( (dbuffer==NULL) || (tbuffer==NULL) || (pbuffer==NULL))
    {
      fprintf(stderr,"Database: problem to allocate memory -> Exiting.\n");
      return(1);
    }

  /* Connect to database */
  if (Connect(data))
    {
      return(2);
    }

#ifdef DEBUG_DB
  printf("database> After connect.\n");
#endif

  snprintf(select_ptr,MAX_QUERY_LENGTH,"SELECT full_payload FROM schema"); 
  if (Select(select_ptr,data) == 0)
    {
      /* Database is not able to store full payload.
       * Print a warning message and a hint how to
       * activate it. Additionally we have to disable
       * the feature and go on (or should we exit?)
       */
      fprintf(stderr,"The database doe not contain the full payload.\n");
      return(1);
    }

  return(0);
}

/* Function: Connect(GPoptions * data)
 *
 * Purpose: Database independent function to initiate a database 
 *          connection
 */

int 
Connect(GPoptions * data)
{
#ifdef MYSQL_DB  
  int x; 
  if ( data->DBtype == DB_MYSQL) 
    {
      data->m_sock = mysql_init(NULL);
      if(data->m_sock == NULL)
	{
	  fprintf(stderr,"database: Connection to database '%s' failed\n", 
		 data->DBname);
	  return(2);
	}
      x = 0;
      if( mysql_real_connect(data->m_sock,"localhost", data->DBuser,
	       data->DBpassword, data->DBname, x, data->SocketName, 0) == 0)
	{
	  if(mysql_errno(data->m_sock))
	    {
	      fprintf(stderr, "database: mysql_error: %s\n", 
		      mysql_error(data->m_sock));
	      return(1);
	    }
	  fprintf(stderr,"database: Failed to logon to database '%s'\n", 
		  data->DBname);
	  return(1);
	}
    }
#endif /* MYSQL */
#ifdef POSTGRES
    if ( data->DBtype == DB_POSTGRES ) 
      {
	data->p_sock=PQsetdbLogin(NULL, data->SocketName, NULL, NULL,
				 data->DBname, data->DBuser, data->DBpassword);
	if (PQstatus(data->p_sock) == CONNECTION_BAD)
	  {
	    PQfinish(data->p_sock);
	    if (PQerrorMessage(data->p_sock))
	      fprintf(stderr, "database: psql_error: %s\n",
		     PQerrorMessage(data->p_sock));
	    fprintf(stderr, "database: Failed to logon to postgresql database "
		    "'%s'\n", data->DBname);
	    return(1);
	  }
      }
#endif /* POSTGRES */
    return(0);
}


/* Function: Disconnect(GPoptions * data)
 *
 * Purpose: Close connection to database and free 
 *          allocated memory.
 */

void 
Disconnect(GPoptions * data)
{

#ifdef DEBUG_DB
  printf("Database> Closing connection to database \"%s\"\n", data->dbname);
#endif
  if (data)
    {
#ifdef MYSQL_DB
      if((data->DBtype==DB_MYSQL) && data->m_sock ) 
	mysql_close(data->m_sock);
#endif 
#ifdef POSTGRES
      if ( (data->DBtype==DB_POSTGRES) && data->p_sock) 
	PQfinish(data->p_sock);
#endif
    }

  /* to be kindly we free some memory... (but exit() will do it too... */
  free(insert_ptr);
  free(select_ptr);
}

/* Function: Insert(char * query, GPoptions * data)
 *
 * Purpose: Database independent function for SQL inserts
 * 
 * Arguments: query (An SQL insert)
 *
 * Returns: 1 if successful, 0 if fail
 */

int 
Insert(char * query, GPoptions * data)
{
    int result = 0;

#ifdef MYSQL_DB
    if ( data->DBtype == DB_MYSQL ) 
      {
        if(!(mysql_query(data->m_sock,query)))
        {
            result = 1;
        }
        else
        {
            if(mysql_errno(data->m_sock))
            {
              fprintf(stderr,"database: mysql_error: %s\nSQL=%s\n", 
                           mysql_error(data->m_sock), query);
            }
        }
      }
#endif /* MYSQL */

#ifdef POSTGRES
    if ( data->DBtype == DB_POSTGRES ) 
      {
	data->p_result = PQexec(data->p_sock,query);
	if (PQresultStatus(data->p_result) == PGRES_COMMAND_OK)
	    {
	      result = 1;
	    }
	else
	  {
	    if (PQerrorMessage(data->p_sock))
	      fprintf(stderr,"database: psql_error: %s\nSQL=%s\n",
		   PQerrorMessage(data->p_sock),query);
	  }

	/* clear result to avoid memory leak */
	PQclear(data->p_result);
      }
#endif /* POSTGRES */
#ifdef DEBUG_DB
    if(result)
        printf("Database> (%s) executed\n", query);
    else
        printf("Database> (%s) failed\n", query);
#endif

    return result;
}


/* Function: Select(char * query, GPoptions * data)
 *
 * Purpose: Database independent function for SQL selects that 
 *          return a non zero int
 * 
 * Arguments: query (An SQL insert)
 *
 * Returns: result of query if successful, 0 if fail
 */

int 
Select(char * query, GPoptions * data)
{
    int result = 0;

#ifdef MYSQL_DB
    if ( data->DBtype == DB_MYSQL ) 
      {
        if(mysql_query(data->m_sock,query))
	  {
            result = 0;
	  }
        else
	  {
            if(!(data->m_result = mysql_use_result(data->m_sock)))
	      {
                result = 0;
	      }
            else
	      {
                if((data->m_row = mysql_fetch_row(data->m_result)))
		  {
                    if(data->m_row[0] != NULL)
		      {
                        result = atoi(data->m_row[0]);
		      }
		  }
	      }
            mysql_free_result(data->m_result);
	  }
        if(!result)
	  {
            if(mysql_errno(data->m_sock))
	      {
                fprintf(stderr,"database: mysql_error: %s\n", 
		       mysql_error(data->m_sock));
	      }
	  }
      }
#endif /* MYSQL */
#ifdef POSTGRES
    if ( data->DBtype == DB_POSTGRES ) 
      {
	data->p_result = PQexec(data->p_sock,query);
	if ( PQresultStatus(data->p_result) == PGRES_TUPLES_OK)
	  {
	    if (PQntuples(data->p_result))
	      result = atoi(PQgetvalue(data->p_result,0,0));
#ifdef DEBUG_DB
	printf("Status: %i\n",result);
#endif
	  }
	else
	  {
	    result=0;
	    if (PQerrorMessage(data->p_sock))
	     fprintf(stderr, "database: psql_error: %s\nSQL=%s\n",
		     PQerrorMessage(data->p_sock),query);
	  }

	/* clear result to avoid memory leak */
	PQclear(data->p_result);
      }
#endif /* POSTGRES */
#ifdef DEBUG_DB
    if(result)
        printf("Database> (%s) returned %u\n", query, result);
    else
        printf("Database> (%s) failed\n", query);
#endif

    return result;
}

/* Function: Select1(char * query, GPoptions * data)
 *
 * Purpose: Database independent function for SQL selects that 
 *          return a TEXT string
 * 
 * Arguments: query (An SQL insert)
 *
 * Returns: result of query.
 */

char *
Select1(char * query, GPoptions * data)
{
    int result = 0;
    char *res;

#ifdef MYSQL_DB
    if ( data->DBtype == DB_MYSQL ) 
      {
        if(mysql_query(data->m_sock,query))
	  {
            result = 0;
	  }
        else
	  {
            if(!(data->m_result = mysql_use_result(data->m_sock)))
	      {
                result = 0;
	      }
            else
	      {
                if((data->m_row = mysql_fetch_row(data->m_result)))
		  {
                    if(data->m_row[0] != NULL)
		      {
                        res = (char *) data->m_row[0];
			if (res==NULL) result=0;
		      }
		  }
	      }
            mysql_free_result(data->m_result);
	  }
        if(!result)
	  {
            if(mysql_errno(data->m_sock))
	      {
                fprintf(stderr, "database: mysql_error: %s\n", 
		       mysql_error(data->m_sock));
	      }
	  }
      }
#endif /* MYSQL */
#ifdef POSTGRES
    if ( data->DBtype == DB_POSTGRES ) 
      {
	data->p_result = PQexec(data->p_sock,query);
	if ( PQresultStatus(data->p_result) == PGRES_TUPLES_OK)
	  {
	    if (PQntuples(data->p_result))
              {
		res = (char *) PQgetvalue(data->p_result,0,0);
	      }

#ifdef DEBUG_DB
	printf("Status: %s\n",res);
#endif
	  }
	else
	  {
	    result=0;
	    if (PQerrorMessage(data->p_sock))
	      fprintf(stderr,"database: psql_error: %s\nSQL=%s\n",
		     PQerrorMessage(data->p_sock),query);
	  }

	/* clear result to avoid memory leak */
	PQclear(data->p_result);
      }
#endif /* POSTGRES */

    return res;
}


/* Function: ReadConfig()
 *
 * Purpose: read configuration files to set at least the parameter
 *          for database access (avoid password on command line)
 *          spaces/tabs at the beginning are removed, spaces at
 *          the end too. But maybe we should use a better parser...
 *
 * Author: DG
 * 
 */

void 
ReadConfig(GPoptions* Options)
{
  FILE *fd = NULL;
  char *token =NULL;
  char *arg=NULL;
  char *val=NULL;
  char line[STRLEN];
  int linenumber=0;
  int i,len;

  if ( (fd=fopen(Options->ConfigFile,"r")) == NULL ) 
    {
      fprintf(stderr, "Can not open config file \"%s\": %s\n", 
	      Options->ConfigFile, strerror(errno));
      exit(1);
    }
  while((fgets(line, STRLEN, fd))!=NULL ) 
    { 
      linenumber++;

      /* remove end of line */
      if ( (token = strchr (line,'\n')) ) *token='\0';

      /* skip if line starts with a comment: if no comment sign */
      /* is found we are sure that we can derefence the pointer */
      /* before any comment sign!*/

      if ( line[0]=='#' ) continue;

      /* remove any comments at end of line and if a comment */
      /* sign is escaped: remove backslash from string!      */
      token = line;
      while ( (token = strchr(token,'#')) && (*(token-1)=='\\' ) )
      {
         val=token;
         arg=token-1;
	 while (*arg) {
		 *arg=*token;
		 arg++;
		 token++;
	 }
	 token=val;
      }

      /* Maybe token is pointing now to a real comment sign */
      if (token && *token=='#') *token='\0';

      /* next if line empty (or only a comment) */
      if (line[0]=='\0') continue;

      /* Up to ":" or "=" we get the argument */
      if ( (strstr(line,":") == NULL) && (strstr(line,"=") == NULL) )
        {
           fprintf(stderr,"\nIgnoring unknown entry in config file: \n");
           fprintf(stderr,"\n> %s (missing : or = ?)\n\n",line);
           continue;
        }

      arg = strtok(line,":=");
      /* second argument is the value, remove tabs */
      val = strtok(NULL,"\0");
      /* remove leading spaces or tabs */
      while ( arg[0]==' ' || arg[0]=='\t') arg++;
      while ( val[0]==' ' || val[0]=='\t') val++;
      /* remove spaces at end of value */
      while ( (strlen(arg) > 1) && (arg[strlen(arg)-1] == ' ') )
		    arg[strlen(arg)-1]='\0';
      while ( (strlen(val) > 1) && (val[strlen(val)-1] == ' ') )
		    val[strlen(val)-1]='\0';
#ifdef DEBUG_READCONFIG
      printf("Line: %i Argument: <%s> Value: <%s>\n",linenumber,arg,val);
#endif
      ArgCheck(val);

      len=strlen(val);
      if(len > STRLEN-2) 
	{
	  fprintf(stderr, "Value in config file too long: %s\n",val);
	  exit(1);
	}

      if ( !strncasecmp(arg,"dbuser",6) )
        {
           strncpy(Options->DBuser,val,len);
	   Options->DBuser[len]='\0';
	   continue;
        } 

      if ( !strncasecmp(arg,"dbpassword",10) )
	{
	      strncpy(Options->DBpassword,val,len);
	      Options->DBpassword[len]='\0';
	      continue;
	}

      if ( !strncasecmp(arg,"dbname",6) )
	{
	      strncpy(Options->DBname,val,strlen(val));
	      Options->DBname[len]='\0';
	      continue;
	}

      if ( !strncasecmp(arg,"socketname",10) )
	{
	  if (len > UNIX_PATH_MAX)
	    {
	      fprintf(stderr,"Length of socketname too large!\n");
	      fprintf(stderr,"Maximum length: %u\n",UNIX_PATH_MAX);
	      fprintf(stderr,"Actual length : %u\n",len);
	      fprintf(stderr,"Exiting.\n");
	      exit(1);
	    }
	  strncpy(Options->SocketName,val,len);
	  Options->SocketName[len]='\0';
	  continue;
	}

      if ( !strncasecmp(arg,"dbtype",7) )
	{
#ifdef MYSQL_DB
	  if (!strncasecmp(val,"MYSQL",5))
	    { 
#ifdef DEBUG_DB
	      printf("Database type: MySQL\n");
#endif
	      Options->DBtype=DB_MYSQL;
	    }
#endif
#ifdef POSTGRES
	  if (!strncasecmp(val,"Postgres",8))
	    { 
#ifdef DEBUG_DB
	      printf("Database type: Postgres\n");
#endif
	      Options->DBtype=DB_POSTGRES ;
	    }
#endif
	  continue;
	}

      /* print warning to stderr to avoid conflicts with stdout */
      fprintf(stderr, "WARNING: unknown option in line %i: %s\n", linenumber,
	      arg);
      
    }

  fclose(fd);
}


/* 
 *  Function: ArgCheck
 *
 *  Purpose: Check for occurence of wrapper around the arguments
 *           in the config file. We remove ' and " if they are at
 *           the beginning and the end.
 *
 *  Arguments: the extracted value as string
 *
 *  Author: DG
 *
 */

void 
ArgCheck(char *arg)
{
  char Value[STRLEN];
  int len;
  
  len=strlen(arg);

  if (len < 2) return;
#ifdef DEBUG_READCONFIG
  printf("Arg: %s, Len: %i\n",arg,len);  
#endif

  if ( ((index(arg,'"')==arg) && (rindex(arg,'"')==arg+len-1)) ||
       ((index(arg,'\'')==arg) && (rindex(arg,'\'')==arg+len-1)) )
   {
       strncpy(Value,arg+1,len-2);
       /* dont't forget to end the string... */
       Value[len-2]='\0';
#ifdef DEBUG_READCONFIG
       printf("Value: %s\n",Value);
#endif
       strcpy(arg,Value);
   }
  return;
}


/* 
 * Function: GetPacket
 *
 * Purpose: Fetches the information to rebuild a network pcap packet
 *          from the database based on sid and cid
 *
 * Arguments: GPoptions, containing database information, sid and cid
 *
 * Returns: 0 on success, 1 on error
 *
 * Author: DG
 *
 */

int
GetPacket(GPoptions *Options)
{
  char *result;

  /* First get the encoding */
  snprintf(select_ptr,MAX_QUERY_LENGTH,"SELECT encoding FROM sensor "
	   "WHERE sid = '%i'",Options->sid);
 
  Options->DBencoding = Select(select_ptr,Options);
#ifdef DEBUG_GP
  printf("Encoding: %i\n",Options->DBencoding);
#endif
  
  if (Options->DBencoding != ENCODING_HEX)
    {
      fprintf(stderr,"Actually only HEX encoding is supported\n");
      return(1);
    }
  
  /* Next get the pcap_header */
  snprintf(select_ptr, MAX_QUERY_LENGTH, "SELECT pcap_header FROM data "
	   "WHERE sid = '%i' AND cid = '%i'",Options->sid,Options->cid);
 
  result = Select1(select_ptr,Options);
  if (result[0]=='\0') 
    {
      /* We don't have a pcap_header, so skip rebuilding */
      fprintf(stderr,"No pcap header, we can't rebuild the network packet.\n");
      return(1);
    }

  if (strlen(result) > (sizeof(struct pcap_pkthdr)*2))
    {
      fprintf(stderr,"Error in pcap_header, answer is too large: %i!\n",
	     strlen(result));
      return(1);
    }
#ifdef DEBUG
  printf("result: <%s>, len %i\n",result,strlen(result));
#endif
  snprintf(pbuffer,(sizeof(struct pcap_pkthdr)*2+1),"%s",result);

  /* Next get the full header of the network packet */
   snprintf(select_ptr, MAX_QUERY_LENGTH, "SELECT data_header FROM "
	    "data WHERE sid = '%i' AND cid = '%i'",Options->sid,Options->cid);
 
  result = Select1(select_ptr,Options);

  if (result[0]=='\0') 
    {
      /* We don't have a pcap_header, so skip rebuilding */
      fprintf(stderr,"No full header, we can't rebuild the network packet.\n");
      return(1);
    }

  if (strlen(result) > MAX_QUERY_LENGTH)
    {
      fprintf(stderr,"Please increase 2 * MAX_QUERY_LENGTH to at least %i\n",
	     strlen(result));
      return(1);
    }

#ifdef DEBUG_GP
  printf("result: <%s>, len %i\n",result,strlen(result));
#endif

  snprintf(tbuffer,MAX_QUERY_LENGTH*2,"%s",result);
  
  /* Finally get the payload of the packet */

  snprintf(select_ptr, MAX_QUERY_LENGTH, "SELECT data_payload FROM data WHERE sid = '%i' AND cid = '%i'",Options->sid,Options->cid);
  
  result = Select1(select_ptr,Options);
  if (result[0]=='\0') 
    {
      fprintf (stderr,"Warning: No high layer payload.\n");
    }

  /* Final test, should never happen */
  if (strlen(result) > MAXBUFFSIZE*2)
    {
      fprintf(stderr,"Please increase 2 * MAXBUFFSIZE to at least %i\n",
	     strlen(result));
      return(1);
    }
#ifdef DEBUG_GP
  printf("result: <%s>, len %i\n",result,strlen(result));
#endif
  snprintf(dbuffer,MAXBUFFSIZE*2,"%s",result);

  return(0);
}


/* 
 * Function: BuildPacket
 *
 * Purpose: builds a network pcap packet and writes it to a file
 *
 * Arguments: GPoptions, containing database information, sid and cid
 *
 * Returns: 0 on success, 1 on error
 *
 * Author: DG
 *
 */

int 
BuildPacket(GPoptions* Options)
{
  FILE *fd=NULL;
  int snaplen;
  struct pcap_file_header hdr;
  struct pcap_pkthdr phdr;
  char tv[9];
  static char packet[MAXBUFFSIZE];
  char * packet_ptr;
  char *ptr;
  int i,j;
  long l;

  /* First open the file if possible or stdout */

  if (Options->DumpFile[0]=='-' && Options->DumpFile[1]=='\0')
    {
      /* Dump to stdout */
      fd=stdout;
    }
  else
    {
      fd=fopen(Options->DumpFile,"w");
    }
     
  /* Check file descriptor */
  if ( fd == NULL ) 
    {
      fprintf(stderr, "Can not open pacp file for writing \"%s\": %s\n", 
	     Options->DumpFile, strerror(errno));
      return(1);
    }

  /* snaplen is len of payload plus header */
  /* for HEX we have to divide by two -> two HEX characters
   * represent one binary byte.
   */

  snaplen=(strlen(tbuffer)+strlen(dbuffer))/2;

  /* Build pcap file header */
  hdr.magic=0xa1b2c3d4;
  hdr.version_major= PCAP_VERSION_MAJOR;
  hdr.version_minor= PCAP_VERSION_MINOR;
  hdr.thiszone=0;                        /* No offset due to timezone */
  hdr.snaplen=snaplen;
  hdr.sigfigs=0;
  hdr.linktype=DLT_EN10MB;

  /* write the pcap file header */
  if (fwrite((char*) &hdr, sizeof(hdr),1,fd) != 1)
    {
      fprintf(stderr,"Problem writing header file.\n");
      fclose(fd);
      return(1);
    }
 
  /* Limit to 32 bit */
  tv[8]='\0'; 

  /* create pcap packet header, this consists of 4 fields a 4 byte */
  /* We have to take care of endianess! */
 
  memcpy(tv,pbuffer,8);
  pbuffer+=8;
  phdr.ts.tv_sec=htonl(strtol(tv,(char **)NULL,16));

  memcpy(tv,pbuffer,8);
  pbuffer+=8;
  phdr.ts.tv_usec=htonl(strtol(tv,(char **)NULL,16));

  memcpy(tv,pbuffer,8);
  pbuffer+=8;
  phdr.caplen=htonl(strtol(tv,(char **)NULL,16));

  memcpy(tv,pbuffer,8);
  pbuffer+=8;
  phdr.len=htonl(strtol(tv,(char **)NULL,16));

#ifdef DEBUG_BP
  printf("len, caplen: %i, %i\n",phdr.len,phdr.caplen);
  printf("snaplen    : %i\n",snaplen);  
#endif 

  if (snaplen != phdr.caplen)
    {
      fprintf(stderr,"WARNING: caplen and snaplen different: %i / %i",
	      snaplen,phdr.caplen);
    }

  /* write the pcap  header */
  if (fwrite((char*) &phdr, sizeof(phdr),1,fd) != 1)
    {
      fprintf(stderr,"Problem writing header file.\n");
      fclose(fd);
      return(1);
    }
 
  tv[2]='\0';
  i=0;
  j=strlen(tbuffer);
  packet_ptr=packet;

  /* copy header to packet, change HEX to BIN */
  while (i<j)
    {
      memcpy(tv,tbuffer,2);
      packet_ptr[0]=(char )strtol(tv,(char **)NULL,16);
      packet_ptr++; 
      tbuffer+=2; 
      i+=2;
    }

  i=0;
  j=strlen(dbuffer);

  /* copy header to packet, change HEX to BIN */
  while (i<j)
    {
      memcpy(tv,dbuffer,2);
      packet_ptr[0]=(char )strtol(tv,(char **)NULL,16);
      packet_ptr++; 
      dbuffer+=2; 
      i+=2;
    }

  /* write the packet to the file */
  if (fwrite(packet, snaplen, 1, fd) != 1)
    {
      fprintf(stderr,"Problem writing header file.\n");
      fclose(fd);
      return(1);
    }

#ifdef DEGUG_BP 
  printf("Done.\n");
#endif

  fclose(fd);
  return(0);
}
